/**************************************************************/
/*                                                            */
/*  Program to help convert a CICS COBOL program to be ECI    */
/*   callable.  The program will attempt to remove most       */
/*   dependencies on 3270 terminals.                          */
/*                                                            */
/*  Written by Jim MacNair - October, 1997                    */
/*                                                            */
/*  The program produces two or three output files from       */
/*  a single CICS COBOL source program.  The output files     */
/*  are as follows:                                           */
/*                                                            */
/*   1) Converted source program.                             */
/*   2) Message file, including all changes made, and         */
/*      indicating any conversions which could not be         */
/*      handled automatically, as well as any errors          */
/*      encountered.                                          */
/*   3) Optionally, a new COBOL 3270 front end program.       */
/*      This program is intended to drive the newly           */
/*      converted program.  This continues to allow           */
/*      3270 access to the converted program while            */
/*      maintaining only a single version of the              */
/*      program,                                              */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/* Change History                                             */
/*                                                            */
/* V1.2                                                       */
/*                                                            */
/* 01/07/99 - Fixed bug with ERASESW invalid character        */
/* 01/07/99 - Changed handling of SEND parameters             */
/* 01/07/99 - Added support for -INC statements (includes)    */
/* 01/07/99 - Added support for ASSIGN TERMCODE               */
/* 01/07/99 - Fixed bug with COPY at end of variable name     */
/* 01/07/99 - Fixed bug with END PROGRAM and generation of    */
/*            BADMAPNAME paragraph.                           */
/* 01/07/99 - Write message for LINK and XCTL statements      */
/* 01/07/99 - Handle -INC as COPY.                            */
/* 01/07/99 - Write messages for any LINK or XCTL statements. */
/* 01/15/99 - Set USRCALEN to 0 on simple return              */
/*                                                            */
/* V1.1                                                       */
/*                                                            */
/* 10/12/98 - Added support for ALARM on EXEC CICS SEND       */
/* 10/15/98 - Added support for statistics/counters           */
/* 10/16/98 - Added support for Copy books                    */
/* 10/16/98 - Corrected bug with multiple keywords in line    */
/*                                                            */
/* V1.02                                                      */
/*                                                            */
/* 5/21/97 - Improved search for EXEC CICS commands           */
/*                                                            */
/* V1.01                                                      */
/*                                                            */
/*  4/1/98 - Fixed bug where the program was confused by      */
/*   a CICS parameter string in a variable name.              */
/*                                                            */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Required include for C functions                           */
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define MAX_NAME_LEN   32
#define MAX_LINE_LEN   255
#define MAX_PROG_NAME  30
#define COBOL_FILE_EXT ".ccp"
#define MAX_COPY_DEPTH 10
#define PROTOTYP_FILE  "NEW3270.TXT"
#define INS_SEND       "INSERT BMS SEND HERE"
#define INS_RECEIVE    "INSERT BMS RECEIVE HERE"
#define LINK_NAME_STR  "&NEWPROG;"
#define BAD_FRONTEND_FILE "\n Unable to open new 3270 front end file\n"
#define CREATE_NEW_FRONT_END "\n Creating new 3270 front end prototype file - %s\n"

/* EXEC CICS keywords */
#define EXEC_STR      "EXEC"
#define CICS_STR      "CICS"
#define RECEIVE_STR   "RECEIVE"
#define SEND_STR      "SEND"
#define CONVERSE_STR  "CONVERSE"
#define MAP_STR       "MAP"
#define TEXT_STR      "TEXT"
#define CONTROL_STR   "CONTROL"
#define ACCUM_STR     "ACCUM"
#define CONVID_STR    "CONVID"
#define SESSION_STR   "SESSION"
#define BUFFER_STR    "BUFFER"
#define PAGE_STR      "PAGE"
#define INTO_STR      "INTO"
#define FROM_STR      "FROM"
#define CTLCHAR_STR   "CTLCHAR"
#define MAPSET_STR    "MAPSET"
#define MAPONLY_STR   "MAPONLY"
#define DATAONLY_STR  "DATAONLY"
#define FREEKB_STR    "FREEKB"
#define FRSET_STR     "FRSET"
#define ALARM_STR     "ALARM"
#define ROUTE_STR     "ROUTE"
#define WAIT_STR      "WAIT"
#define RETURN_STR    "RETURN"
#define ERASE_STR     "ERASE"
#define ERASEAUP_STR  "ERASEAUP"
#define CURSOR_STR    "CURSOR"
#define TRANSID_STR   "TRANSID"
#define COMMAREA_STR  "COMMAREA"
#define LENGTH_STR    "LENGTH"
#define ADDRESS_STR   "ADDRESS"
#define TCTUA_STR     "TCTUA"
#define HANDLE_STR    "HANDLE"
#define AID_STR       "AID"
#define MAPFAIL_STR   "MAPFAIL"
#define ASSIGN_STR    "ASSIGN"
#define TERMCODE_STR  "TERMCODE"
#define LINK_STR      "LINK"
#define XCTL_STR      "XCTL"
#define ENDEXEC_STR   "END-EXEC"
/* AID Keys keywords */
#define PA_STR        "PA"
#define PF_STR        "PF"
#define CLEAR_STR     "CLEAR"
#define DFHCLEAR_STR  "DFHCLEAR"
#define DFHPF_STR     "DFHPF"
#define DFHPA_STR     "DFHPA"
#define PEN_STR       "PEN"
#define LIGHTPEN_STR  "LIGHTPEN"
#define DFHPEN_STR    "DFHPEN"
#define OPID_STR      "OPID"
#define OPERID_STR    "OPERID"
#define DFHOPID_STR   "DFHOPID"
#define TRIG_STR      "TRIG"
#define TRIGGER_STR   "TRIGGER"
#define DFHTRIG_STR   "DFHTRIG"
/* COBOL Language keywords */
#define PROCEDURE_STR "PROCEDURE"
#define DIVISION_STR  "DIVISION"
#define LINKAGE_STR   "LINKAGE"
#define SECTION_STR   "SECTION"
#define LEVEL01_STR   "01"
#define PROGRAMID_STR "PROGRAM-ID"
#define COPY_STR      "COPY"
#define INC_STR       "-INC"
#define END_STR       "END"
#define PROGRAM_STR   "PROGRAM"
/* EIB variable names and replacement values */
#define EIBTRMID_STR  "EIBTRMID"
#define EIBTRMID_NEW  "USRTRMID"
#define EIBTRNID_STR  "EIBTRNID"
#define EIBTRNID_NEW  "USRTRNID"
#define EIBAID_STR    "EIBAID"
#define EIBAID_NEW    "USRAID"
#define EIBCPOSN_STR  "EIBCPOSN"
#define EIBCPOSN_NEW  "USRCPOSN"
#define EIBCALEN_STR  "EIBCALEN"
#define EIBCALEN_NEW  "USRCALEN"
#define DFHCOMM_NEW   "USRCOMM OF DFHCOMMAREA"
#define DFHCOMM_STR   "DFHCOMMAREA"
/* CICS copybook names */
#define DFHAID_STR    "DFHAID"
#define DFHBMSCA_STR  "DFHBMSCA"
#define DFHEIBLK_STR  "DFHEIBLK"

/************************************************************/
/* Structure definition for BMS command table               */
/************************************************************/
#define MAX_BMS 100
typedef struct _BMSTABLE
{
	char sendRecv;
	char eraseSW;
	char sendType;
	char freeKB;
	char frSet;
	char mapName[9];
	char mapSet[9];
	char alarmSW;
	char dataLen[8];
	char dataArea[80];
}	BMSTABLE;

 /*
  * Copyright and version number constants
  */

static char copyright[] = "(C) Copyright IBM Corp, 1998 ";
static char Version[]=\
"@(#)cblcvrt V1.2 - COBOL 3270 to ECI conversion  - Jim MacNair ("__DATE__")";

 /*
  * Global variables
  */

	FILE	*cobolpgm;    /* input COBOL program             */
    FILE	*newpgm;      /* converted COBOL program         */
    FILE	*msgfile;     /* conversion program messages     */
	FILE	*frontpgm;    /* optional new 3270 front end     */
	FILE	*prototyp;    /* template for new 3270 front end */
	FILE*	saveinput[MAX_COPY_DEPTH];  /* Save input file handles */
	FILE*	saveoutput[MAX_COPY_DEPTH]; /* Save output file handles */
	int		savecopycount[MAX_COPY_DEPTH]; /* Save copy file line count */
	int		savecopyout[MAX_COPY_DEPTH]; /* Save output line counts */
	int		copydepth=0;  /* current copy book depth         */
    int		inCounter=0;  /* input line counter              */
    int		outCounter=0; /* output line counter             */
	int		incopy=0;     /* Copy Book indicator             */
    int		badMap=0;     /* bad map paragraph indicator     */

/*
 * Statement counters
 */
	int	totalExecCICS=0;  /* Total number of EXEC CICS statements */
	int	totalSendText=0;  /* Total number of EXEC CICS SEND TEXT */
	int	totalSendMap=0;   /* Total number of EXEC CICS SEND MAP */
	int	totalControl=0;   /* Total number of EXEC CICS SEND CONTROL */
	int	totalRecvMap=0;   /* Total number of EXEC CICS RECEIVE MAP */
	int	totalConverse=0;  /* Total number of EXEC CICS CONVERSE */
	int	totalHandleAID=0; /* Total number of HANDLE AIDs */
	int	totalMapfail=0;   /* Total number of HANDLE MAPFAIL */
	int	totalCopy=0;      /* Total number of COPY statements */
	int	totalInc=0;       /* Total number of -INC statements */
	int	totalReturn=0;    /* Total number of RETURN statements */
	int	totalAddress=0;   /* Total number of ADDRESS statements */
	int	totalAlarm=0;     /* Total number of ALARM parameters */
	int	totalMaponly=0;   /* Total number of MAPONLY parameters */
	int	totalDataonly=0;  /* Total number of DATAONLY parameters */
	int	totalErase=0;     /* Total number of ERASE parameters */
	int	totalEraseAUP=0;  /* Total number of ERASEAUP parameters */
	int	totalAccum=0;     /* Total number of ACCUM parameters */
	int	totalPage=0;      /* Total number of PAGE parameters */
	int	totalConvid=0;    /* Total number of APPC SENDS (CONVID or SESSION) */
	int	totalSendTerm=0;  /* Total number of SENDs with FROM or CTLCHAR only */
	int	totalSendNoparm=0; /* Total number of SENDs with no other parameters */
	int	totalBuffer=0;    /* Total number of RECEIVE BUFFER */
	int	totalFreekb=0;    /* Total number of FREEKB parameters */
	int	totalFrset=0;     /* Total number of FRSET parameters */
	int	totalRoute=0;     /* Total number of ROUTE statements */
	int totalTermcode=0;  /* Total number of ASSIGN TERMCODE statements */
	int totalLink=0;      /* Total number of LINK statements */
	int totalXctl=0;      /* Total number of XCTL statements */
	int	totEibaid=0;      /* Total number of EIBAID references */
	int	totEibtrmid=0;    /* Total number of EIBTRMID references */
	int	totEibtrnid=0;    /* Total number of EIBTRNID references */
	int	totEibcalen=0;    /* Total number of EIBCALEN references */
	int	totEibcposn=0;    /* Total number of EIBCPOSN references */
	int	totCommarea=0;    /* Total number of DFHCOMMAREA references */
	int totSendRecv=0;    /* Total number of SEND and RECEIVE statements */

/*
 * This routine will restore the input and
 * output files after a copy book has been
 * processed.
 */

void restoreFiles()

{
	int		i;

	/* restore the input and output files */
	cobolpgm = saveinput[0];
	newpgm = saveoutput[0];
	inCounter = savecopycount[0];
	outCounter = savecopyout[0];

	for (i=0; i < MAX_COPY_DEPTH - 1; i++)
	{
		saveinput[i] = saveinput[i+1];
		saveoutput[i] = saveoutput[i+1];
		savecopycount[i] = savecopycount[i+1];
		savecopyout[i] = savecopyout[i+1];
	}

	/* Clear the last saved file */
	saveinput[MAX_COPY_DEPTH - 1]  = NULL;
	saveoutput[MAX_COPY_DEPTH - 1] = NULL;
	savecopycount[MAX_COPY_DEPTH - 1] = 0;
	savecopyout[MAX_COPY_DEPTH - 1] = 0;

	/* decrement the copy book depth */
	copydepth--;
}

/*
 * This routine will switch the input and output
 * files to a copy book.  It will basically use
 * a push down stack to remember the current 
 * input and output files.
 */

void openCopyBook(char *filename, char *cobolLine, int *dontwrite)

{
	int		i;
	char*	cptr;
	char	pgmfilename[255];

	/* ignore any copy for CICS copybooks */
	if ((strcmp(filename, DFHAID_STR) == 0) ||
		(strcmp(filename, DFHBMSCA_STR) == 0) ||
		(strcmp(filename, DFHEIBLK_STR) == 0))
	{
		return;
	}

	/* save the current files and copy counter */
	for (i=MAX_COPY_DEPTH-1; i>0; i--)
	{
		saveinput[i]  = saveinput[i - 1];
		saveoutput[i] = saveoutput[i - 1];
		savecopycount[i] = savecopycount[i-1];
		savecopyout[i] = savecopyout[i-1];
	}

	/* copy the current files into the first saved entry */
	saveinput[0] = cobolpgm;
	saveoutput[0] = newpgm;
	savecopycount[0] = inCounter;
	savecopyout[0] = outCounter;

	/* Increase the copy book depth */
	copydepth++;

	/* Tell the user what we are doing */
	fprintf(msgfile, "  opening copy book %s at line %d\n", filename, inCounter);

	/* Clear the existing file handles */
	cobolpgm = NULL;
	newpgm = NULL;
	inCounter = 0;
	outCounter = 0;

	/* try to open the new copy book */
	if ((cobolpgm = fopen(filename, "r")) == NULL)
	{
		/* didn't work */
		fprintf(msgfile, " ***error*** opening copy file %s\n", filename);
		restoreFiles();
	}
	else
	{
		/* get the copy book file name */
		strcpy(pgmfilename, filename);

		/* check for an existing file extension */
		cptr = strchr(pgmfilename, '.');
		if (cptr != NULL)
		{
			cptr[0] = '\0';
		}

		/* change the file extension to new */
		strcat(pgmfilename, ".new");

		/* try to open the new file */
		if ((newpgm = fopen(pgmfilename, "w")) == NULL)
		{
			/* failed - give up on the copy book */
			fclose(cobolpgm);
			fprintf(msgfile," ***error*** opening output copy file %s\n", pgmfilename);
			restoreFiles();
		}
		else
		{
			/* Write the copy book statement out to the current file first */
			fprintf(saveoutput[0], "%-80s\n", cobolLine);

			/* Count the output lines */
			(savecopyout[0])++;

			/* Do not write the copy statement to the new file */
			*dontwrite = 1;
		}
	}
}

/*
 * This routine skips blanks in a string
 */

char * skipBlanks(char *str)

{
   while (*str == ' ')
   {
      str++;
   } /* endwhile */

   return str;
}

/*
 * This routine finds the value in a parameter from an
 * EXEC CICS statement.  The value is enclosed within
 * parentheses.
 */

void getValue(const char *map, char *mapName)

{
	char *beginVal;
	char *endVal;
	char tempval[MAX_LINE_LEN];

	*(mapName) = '\0';  /* set parameter value to null */

	if (map != NULL)
	{
		/* get a copy of the input data */
		memcpy(tempval, map, sizeof(tempval));

		/* search for begin parentheses */
		beginVal = strchr(tempval, '('); 
		if (beginVal != NULL)
		{
			 /* move past begin parentheses */
			beginVal++; 

			/* find end parentheses */
			endVal = strchr(beginVal, ')');

			/* Did we find the end parenthese? */
			if (endVal != NULL)
			{
				/* terminate the string */
				*endVal = '\0'; 

				/* Return the value */
				strcpy(mapName,beginVal);
			}
		}
	}
}

/*
 * This routine writes a line to the message file
 * that is produced with the new program.
 */

void writeErrorLine(int count, char *msg)
{
	int i;		/* work variable */

	/* offset any copy book lines */
	if (copydepth > 0)
	{
		for (i=0; i<copydepth; i++)
		{
			fprintf(msgfile, " ");
		}
	}

	/* print the message */
	fprintf(msgfile,"line %.4d - %s\n",count + 1,msg);
}

/*******************************************************************/
/* Routine to extract a program name from a full file name.        */
/* Directory and file extension information will be stripped.      */
/*******************************************************************/

void processFilename(char *newFileName, char *newProgName)
{
	char *tempptr;
	char *fileExt;

	/* check for any directories in the file name */
	tempptr = newFileName;
	fileExt = strchr(tempptr,'\\');

	/* check for back slash characters */
	while (fileExt != NULL)
	{
		/* point to character after the directory separator character */
		/*fileExt++;                                                  */
		tempptr = ++fileExt;

		/* check for more sub-directories */
		fileExt = strchr(tempptr, '\\');
	}

	/* Make a copy of the new file name parameter       */
	/*  being sure not to overflow the newFileName area */
	strncpy(newProgName, tempptr, MAX_PROG_NAME);
	newProgName[MAX_PROG_NAME - 1] = '\0';

	/* Isolate the file name without the extension */
	/* Search for a period                         */
	fileExt = strchr(newProgName, '.');

	/* Is there a file extension? */
	if (fileExt != NULL)
	{
		/* Remove the file extension */
		fileExt[0] = '\0';
	}

	/* Check if we found a program id */
	if (strlen(newProgName) > 0)
	{
		/* Translate to Upper Case */
		strupr(newProgName);
	}
}

/*
 * Routine to check for CICS keywords.  This routine
 * is needed to not be confused by keyword strings
 * in variable names.  It will look for a proper
 * delimiting character right after the keyword.
 * Delimiting values can be a period, space or
 * begin parentheses.
 */

char * findKeyword(char *cmdline, const char *keywd)

{
	char *temptr;
	char *prevptr;
	char *endptr;
	char *firstquote;
	char *secondquote;

	temptr = cmdline;
	while (1)
	{
		/* try to find the second string in the first string */
		temptr = strstr(temptr, keywd);

		/* Did we find it? */
		if (temptr == NULL)
		{
			/* Didn't find it = get out */
			break;
		}
		else
		{
			/* Point to the character after the keyword string */
			endptr = temptr + strlen(keywd);

			/* Point to the character before the keyword string */
			prevptr = temptr - 1;

			/* Make sure that the previous and next characters are delimiters */
			if (((endptr[0] == ' ') || (endptr[0] == '.') ||
				 (endptr[0] == '(') || (endptr[0] == '\0')) && 
				 ((prevptr[0] == ' ') || (temptr == cmdline)))
			{
				/* Lastly, check if we are in between quotes */
				firstquote = strchr(cmdline, '\'');
				/* Did we find a quote? */
				if (firstquote != NULL)
				{
					/* Yes, try to find a second quote */
					secondquote = strchr(firstquote + 1, '\'');
				}

				/* check if the key word is found in a literal */
				if ((firstquote == NULL) || (secondquote == NULL) ||
					(temptr < firstquote) || (temptr > secondquote))
				{
					/* We are not in a literal */
					break;
				}
				else
				{
					/* skip to the end of the second quote */
					temptr = secondquote + 1;
				}
			}
			else
			{
				/* move past the current string */
				temptr++;
			}
		}
	}

	/* Return the pointer value */
	return temptr;
}

/*******************************************************************/
/* Routine to check for duplicate BMS commands and only insert     */
/* one into the new 3270 front end program.                        */
/* If a duplicate is found, then the program returns a 1,          */
/* otherwise it returns a 0.                                       */
/*******************************************************************/

int checkDup(BMSTABLE * BMStab, int max)
{
	int rc = 0;
	int i = 0;

	while ((i < max) && (BMStab[i].sendRecv != 0))
	{
		/* check for identical entries in the table */
		if (memcmp(BMStab + i, BMStab + max, sizeof(BMSTABLE)) == 0)
		{
			rc = 1;
		}

		/* move on to the next entry */
		i++;
	}

	/* return the results */
	return rc;
}

/*******************************************************************/
/* Routine to write a line to the new 3270 front end program.      */
/*******************************************************************/

void writePrototypLine(char *newLine)
{
	/* write the line */
	fprintf(frontpgm, "%-80s\n", newLine);
}

/*******************************************************************/
/* Routine to handle parameters on EXEC CICS SEND                  */
/*******************************************************************/

void procSendParm(const char* sw, char swval, int andsw)

{
	char newLine[80]; /* work area for new COBOL line */

	/* Build the line for the IF statement */
	sprintf(newLine,"               %s IS EQUAL TO \'%c\' ", sw, swval);

	if (andsw == 0)
	{
		/* finish the statement line */
		strcat(newLine, "AND");
	}
	else
	{
		strcat(newLine, "THEN");
	}	/* End if */

	/* write the line to the file */
	writePrototypLine(newLine);
}

/*******************************************************************/
/* Routine to build a new 3270 front end program skeleton, based   */
/* on a special template file.                                     */
/*******************************************************************/

void build3270pgm(BMSTABLE *BMStab, char *frontProgName, char *newProgName)
{
	char cobolLine[MAX_LINE_LEN]; /* Current COBOL template Line   */
	char templine[MAX_LINE_LEN];  /* temporary work area           */
	char newLine[80]; /* work area for new COBOL line    */
	char erase[9];	  /* ERASE parameter                 */
	char sendtype[9]; /* MAPONLY or DATAONLY parameter   */
	char freekb[7];   /* FREEKB parameter                */
	char frset[6];    /* FRSET parameter                 */
	char alarm[6];    /* ALARM parameter                 */
	char *progid;	  /* Pointer to PROGRAM-ID paragraph */
	char *linkname;   /* pointer to EXEC CICS LINK program name */
	int  i;			  /* work variable                   */
	int  dontwrite;   /* variable to suppress writing current line */
	int  recvcount;   /* Count of the number of BMS RECEIVE statements */
	int  sendcount;   /* Count of the number of BMS SEND statements */
	int  senducount;  /* Count of the number of BMS SEND statements */
	int	 sendparms;   /* Counter of BMS SEND parameters */

	while (1)
	{
		/* read the next line in the template source file */
		if (fgets(cobolLine, MAX_LINE_LEN, prototyp) == NULL)
		{
			break;
		}	/* End if */

		/* get rid of the newline character */
		if (cobolLine[strlen(cobolLine) - 1] == '\n')
		{
			cobolLine[strlen(cobolLine) - 1] = '\0';
		}

		/* set the variable to write the current line */
		dontwrite = 0;

		/* Check for PROGRAM-ID paragraph */
		progid = findKeyword(cobolLine, PROGRAMID_STR);
		if (progid != NULL)
		{
			/* Truncate the rest of the line */
			progid[sizeof(PROGRAMID_STR) - 1] = '\0';

			/* Finally, append the new program name */
			strcat(progid, ". ");
			strcat(progid, frontProgName);
			strcat(progid, ".");
		}	/* End if */

		/* Look for the replacement string for the Linked To Program */
		linkname = strstr(cobolLine, LINK_NAME_STR);
		if (linkname != NULL)
		{
			/* Save a copy of the end of the line */
			strcpy(templine, linkname + sizeof(LINK_NAME_STR) - 1);

			/* Truncate the current line */
			linkname[0] = '\0';

			/* Insert the new program name */
			strcat(cobolLine, newProgName);

			/* Append the rest of the line */
			strcat(cobolLine, templine);
		}

		/* Check for BMS RECEIVE insertion point */
		if ((strlen(cobolLine) > 6 + sizeof(INS_RECEIVE)) && (cobolLine[6] == '*') &&
		    (memcmp(&cobolLine[8],INS_RECEIVE,sizeof(INS_RECEIVE)) == 0))
		{
			/* do not write out this line */
			dontwrite = 1;

			/* zero out the RECEIVE counter */
			recvcount = 0;

			i = 0;
			while (BMStab[i].sendRecv != 0)
			{
				/* Check for BMS RECEIVE commands */
				if ((BMStab[i].sendRecv == 'R') && (checkDup(BMStab, i) == 0))
				{
					/* increment the RECEIVE counter */
					recvcount++;

					/* move the map name to the USRMAPNAME field in STDHEAD */
					sprintf(newLine,"           MOVE %s TO USRMAPNAME",
							BMStab[i].mapName);
					/* move the map set name to the USRMAPSET field in STDHEAD */
					sprintf(newLine,"           MOVE %s TO USRMAPSET",
							BMStab[i].mapSet);

					/* write out the first part of the EXEC CICS RECEIVE command */
					sprintf(newLine,"           EXEC CICS RECEIVE MAP(%s) MAPSET(%s)",
							BMStab[i].mapName, BMStab[i].mapSet);

					/* write the line to the file */
					writePrototypLine(newLine);

					/* Check for a INTO parameter */
					if (BMStab[i].dataArea[0] != '\0')
					{
						sprintf(newLine,"                   INTO(%s)",
								BMStab[i].dataArea);

						/* write the line to the file */
						writePrototypLine(newLine);
					}

					/* Insert the END-EXEC statement */
					strcpy(newLine,"               END-EXEC");

					/* write the line to the file */
					writePrototypLine(newLine);
				}	/* End if */

				/* Go on to the next BMS command */
				i++;
			}	/* End while */

			/* Display the number of BMS RECEIVE statements found */
			fprintf(msgfile, "Number of BMS RECEIVE statements found in original Program - %d\n",
					recvcount);

			/* Check if we found more than one RECEIVE */
			if (recvcount > 1)
			{
				/* Found more than one receive - manual changes are needed */
				fprintf(msgfile, "**** Warning - Multiple receives found - manual changes required\n");
			}
		}	/* End if */

		/* Check for BMS SEND insertion point */
		if ((strlen(cobolLine) > 6 + sizeof(INS_SEND)) && (cobolLine[6] == '*') &&
		    (memcmp(&cobolLine[8],INS_SEND,sizeof(INS_SEND)) == 0))
		{
			/* do not write out this line */
			dontwrite = 1;

			/* zero out the SEND counters */
			sendcount = 0;
			senducount = 0;

			i = 0;
			while (BMStab[i].sendRecv != 0)
			{
				/* Count the total number of SEND statements found */
				if (BMStab[i].sendRecv == 'S') 
				{
					sendcount++;
				}

				/* Check for BMS SEND commands */
				if ((BMStab[i].sendRecv == 'S') && (checkDup(BMStab, i) == 0))
				{
					/* increment the unique SEND counter */
					senducount++;

					/* Zero out the Send Parameters counter */
					sendparms = 0;

					/* Check for ERASE or ERASEAUP specified */
					switch (BMStab[i].eraseSW)
					{
						case 'E':	strcpy(erase, ERASE_STR);
									sendparms++;
									break;
						case 'U':	strcpy(erase, ERASEAUP_STR);
									sendparms++;
									break;
						default: erase[0] = '\0';;
					}	/* End switch */

					/* Check for ALARM specified */
					alarm[0] = '\0';
					if (BMStab[i].alarmSW == 'A')
					{
						strcpy(alarm, ALARM_STR);
						sendparms++;
					}	/* End if */

					/* Check for MAPONLY or DATAONLY specified */
					switch (BMStab[i].sendType)
					{
						case 'M': strcpy(sendtype, MAPONLY_STR);
									sendparms++;
									break;
						case 'D': strcpy(sendtype, DATAONLY_STR);
									sendparms++;
									break;
						default: sendtype[0] = '\0';
					}	/* End switch */

					/* check for FREEKB parameter */
					freekb[0] = '\0';
					if (BMStab[i].freeKB == 'F')
					{
						strcpy(freekb, FREEKB_STR);
						sendparms++;
					}	/* End if */

					/* check for FRSET parameter */
					frset[0] = '\0';
					if (BMStab[i].frSet == 'F')
					{
						strcpy(frset, FRSET_STR);
						sendparms++;
					}	/* End if */

					/* Start the IF statement */
					sprintf(newLine,"           IF USRMAPNAME = %s AND",
							BMStab[i].mapName);

					/* write the line to the file */
					writePrototypLine(newLine);

					/* get the next condition - check for ERASE */
					procSendParm("ERASESW", BMStab[i].eraseSW,0);

					/* get the next condition - check for ALARM */
					procSendParm("ALARMSW", BMStab[i].alarmSW,0);

					/* get the next condition - check for MAPONLY or DATAONLY */
					procSendParm("SENDTYPE", BMStab[i].sendType,1);

					sprintf(newLine,"               EXEC CICS SEND MAP(%s) MAPSET(%s)",
							BMStab[i].mapName, BMStab[i].mapSet);

					/* write the line to the file */
					writePrototypLine(newLine);

					/* Check for a FROM and/or LENGTH parameter */
					if ((BMStab[i].dataArea[0] != '\0') || (BMStab[i].dataLen[0] != '\0'))
					{
						/* Check if both parameters were specified */
						if ((BMStab[i].dataArea[0] != '\0') && (BMStab[i].dataLen[0] != '\0'))
						{
							/* Insert the FROM and LENGTH parameters */
							sprintf(newLine,"                   FROM(%s) LENGTH(%s)",
									BMStab[i].dataArea, BMStab[i].dataLen);

						}
						else
						{
							/* Check if FROM parameter was specified */
							if (BMStab[i].dataArea[0] != '\0')
							{
								/* Insert the FROM parameter */
								sprintf(newLine,"                   FROM(%s)",
										BMStab[i].dataArea);
							}
							else
							{
								/* Insert the LENGTH parameter */
								sprintf(newLine,"                   LENGTH(%s)",
										BMStab[i].dataLen);
							}
						}

						/* write the line to the file */
						writePrototypLine(newLine);
					}

					/* Check if ERASE, ERASEAUP, DATAONLY, MAPONLY, FREEKB or FRSET */
					/* was specified on the command                                 */
					if ((erase[0] == '\0') && (sendtype[0] == '\0') && 
						(frset[0] == '\0') && (freekb[0] == '\0'))
					{
						/* None were specified - just write the END-EXEC */
						strcpy(newLine,"                   END-EXEC");
					}
					else
					{
						/* At least one was specified - add them to the command */
						sprintf(newLine,"                   %s %s %s %s END-EXEC",
								erase,
								sendtype,
								freekb,
								frset);
					}

					/* write the line to the file */
					writePrototypLine(newLine);

					/* End the IF statement */
					strcpy(newLine,"           END-IF");

					/* write the line to the file */
					writePrototypLine(newLine);
				}	/* End if - check for SEND in BMS command table */

				/* Go on to the next BMS command */
				i++;
			}	/* End while */

			/* Display the number of BMS SEND statements found */
			fprintf(msgfile, "Number of BMS SEND statements found in original Program -    %d\n",
					sendcount);

			/* Check if the number of unique statements was less */
			if (senducount != sendcount)
			{
				/* Display the number of BMS SEND statements found */
				fprintf(msgfile, "Number of unique BMS SEND statements found              -    %d\n",
						senducount);
			}
		}	/* End if - check for BMS send insertion point */

		/* write line to new program file */
		if (dontwrite == 0)
		{
			writePrototypLine(cobolLine);
		}	/* End if */
	}  /* End while */
}

/*
 * Subroutine to replace EIB variable with a usr
 * value, which will be passed in the commarea.
 */

void processEIB(char *cobolLine,
                const char *EIB,
                const char *newEIB,
				int*	EIBcount)

{
   char *valLoc;            /* Location of value to change */
   char errline[80];
   char saveline[MAX_LINE_LEN];

   valLoc = findKeyword(cobolLine, EIB);
   if (valLoc != NULL)
   {
		/* First, copy the rest of the line to a temporary area */
		strcpy(saveline, valLoc + strlen(EIB));

		/* truncate the rest of the line */
		valLoc[0] = '\0';

		/* Insert the new EIB value */
		strcat(valLoc, newEIB);

		/* Replace the end of the line */
		strcat(valLoc, saveline);

		/* Record the change on the message file */
		sprintf(errline, "%s changed", EIB);
		writeErrorLine(outCounter, errline);

		/* Increment the counter */
		(*EIBcount)++;
   }
}

/*
 * This routine handles any EIB variables that must
 * be converted.  It also handles references to
 * DFHCOMMAREA.
 */

void handleEIB(char *cobolLine)
{
   /* check for use of EIB variables which relate to terminals */
   processEIB(cobolLine, EIBTRMID_STR, EIBTRMID_NEW,&totEibtrmid);
   processEIB(cobolLine, EIBTRNID_STR, EIBTRNID_NEW,&totEibtrnid);
   processEIB(cobolLine, EIBAID_STR, EIBAID_NEW,&totEibaid);
   processEIB(cobolLine, EIBCPOSN_STR, EIBCPOSN_NEW,&totEibcposn);
   processEIB(cobolLine, EIBCALEN_STR, EIBCALEN_NEW,&totEibcalen);
   processEIB(cobolLine, DFHCOMM_STR, DFHCOMM_NEW,&totCommarea);
}

/*
 * This routine saves handle aid keys in a buffer
 */

char * processKEY(char *handleaids, char *keyname, char *aidkey)

{
   char *endval;

   /* skip any leading blanks on parameter */
   aidkey = skipBlanks(aidkey);

   /* check if we have valid syntax */
   if (*aidkey++ == '(')
   {
      endval = aidkey;
      /* find the end of the value */
      while ((*endval != ')') && (*endval != '\0'))
      {
         endval++;
      } /* endwhile */

      if (*endval = ')')
      {
         /* terminate the string */
         *endval = '\0';

         /* go to the next byte */
         endval++;

         /* make sure the value is not too long */
         if (strlen(aidkey) > MAX_NAME_LEN)
         {
            *(aidkey + MAX_NAME_LEN) = '\0';
         } /* endif */

         /* Concatenate the separator character */
         strcat(handleaids, "&");

         /* Concatenate the key name */
         strcat(handleaids, keyname);

         /* Concatenate the separator character */
         strcat(handleaids, "=");

         /* Concatenate the key value */
         strcat(handleaids, aidkey);
      } /* endif */
   } /* endif */

   /* return an updated location in the buffer */
   return endval;
}

/*
 * This routine handles an EXEC HANDLE AID command.
 */

void processHandleAID(char *aid,
                      char *ofsline,
                      char *handleaids,
                      char periodch,
                      int saveCount)

{
   char *aidkey;
   char *endkey;
   char keyname[6];
   char pfval[3];
   int  i;

   /* start after the aid keyword */
   aidkey = aid + sizeof(AID_STR);

   /* calculate the end of the command */
   endkey = aidkey + strlen(aidkey);

   while ((*aidkey != '\0') && (aidkey < endkey))
   {
      /* Skip any intermediate blanks */
      aidkey = skipBlanks(aidkey);

      /* Clear the keyname field */
      memset(keyname, '\0', sizeof(keyname));

      /* check for PF key */
      if (memcmp(aidkey, PF_STR, sizeof(PF_STR) - 1) == 0)
      {
         /* move past PF */
         aidkey += sizeof(PF_STR) - 1;

         /* get the first two bytes of aid key */
         memcpy(keyname, PF_STR, sizeof(PF_STR) - 1);

         if ((*aidkey >= '0') && (*aidkey <= '9'))
         {
            /* save the first digit */
            pfval[0] = *aidkey++;
            pfval[1] = '\0';

            /* check for a second digit */
            if ((*aidkey >= '0') && (*aidkey <= '9'))
            {
               /* save the second digit */
               pfval[1] = *aidkey++;
               pfval[2] = '\0';
            } /* endif */

            /* convert to an integer */
            i = atoi(pfval);

            if ((i > 0) && (i <= 24))
            {
               /* force two byte number */
               if (strlen(pfval) == 1)
               {
                  strcat(keyname, "0");
               } /* endif */

               /* finish the aid key name */
               strcat(keyname, pfval);

               /* add to the aid key table */
               aidkey = processKEY(handleaids,keyname,aidkey);
            }
            else
            {
               writeErrorLine(saveCount, "Invalid PF Key");
            } /* endif */
         } /* endif */
      }

      if (memcmp(aidkey, PA_STR, sizeof(PA_STR) - 1) == 0)
      {
         /* move past PA */
         aidkey += sizeof(PA_STR) - 1;

         /* get the first two bytes of aid key */
         memcpy(keyname, PA_STR, sizeof(PA_STR) - 1);

         if ((*aidkey > '0') && (*aidkey <= '3'))
         {
            keyname[2] = *aidkey;  /* save the character */
			keyname[3] = 0;        /* Terminate the string */
            aidkey++;              /* move to next char  */

            /* add to the aid key table */
            aidkey = processKEY(handleaids,keyname,aidkey);
         }
         else
         {
            writeErrorLine(saveCount, "Invalid PA Key");
         } /* endif */
      }

      if (memcmp(aidkey,CLEAR_STR, sizeof(CLEAR_STR) - 1) == 0)
      {
         /* move past the key name */
         aidkey += sizeof(CLEAR_STR) - 1;

         /* set the key name */
         strcat(keyname, CLEAR_STR);

         /* add to the aid key table */
         aidkey = processKEY(handleaids,keyname,aidkey);
      }

      if (memcmp(aidkey,LIGHTPEN_STR, sizeof(LIGHTPEN_STR) - 1) == 0)
      {
         /* move past the key name */
         aidkey += sizeof(LIGHTPEN_STR) - 1;

         /* set the key name */
         strcat(keyname, PEN_STR);

         /* add to the aid key table */
         aidkey = processKEY(handleaids,keyname,aidkey);
      }

      if (memcmp(aidkey,TRIGGER_STR, sizeof(TRIGGER_STR) - 1) == 0)
      {
         /* move past the key name */
         aidkey += sizeof(TRIGGER_STR) - 1;

         /* set the key name */
         strcat(keyname, TRIG_STR);

         /* add to the aid key table */
         aidkey = processKEY(handleaids,keyname,aidkey);
      }

      if (memcmp(aidkey,OPERID_STR, sizeof(OPERID_STR) - 1) == 0)
      {
         /* move past the key name */
         aidkey += sizeof(OPERID_STR) - 1;

         /* set the key name */
         strcat(keyname, OPID_STR);

         /* add to the aid key table */
         aidkey = processKEY(handleaids,keyname,aidkey);
      } /* endif */

         aidkey++;
   } /* endwhile */
}

/*
 * This routine converts a HANDLE AID parameter name
 * to a CICS DFHBMSCA name.
 */

void cicsAIDname(char *aidname, char *cicsname)

{
   char keyval[10];

   memset(keyval, '\0', sizeof(keyval));

   if (memcmp(aidname, PF_STR, sizeof(PF_STR) - 1) == 0)
   {
      strcpy(cicsname, DFHPF_STR);

      if (*(aidname + sizeof(PF_STR) - 1) == '0')
      {
         keyval[0] = *(aidname + sizeof(PF_STR));
      }
      else
      {
         keyval[0] = *(aidname + sizeof(PF_STR) - 1);
         keyval[1] = *(aidname + sizeof(PF_STR));
      } /* endif */

      strcat(cicsname, keyval);
   } /* endif */

   if (memcmp(aidname, PA_STR, sizeof(PA_STR) - 1) == 0)
   {
      strcpy(cicsname, DFHPA_STR);
      cicsname += sizeof(DFHPA_STR) - 1;
      *(cicsname) = *(aidname + sizeof(PA_STR) - 1);
	  *(cicsname + 1) = '\0';
   } /* endif */

   if (memcmp(aidname, CLEAR_STR, sizeof(CLEAR_STR) - 1) == 0)
   {
      strcpy(cicsname, DFHCLEAR_STR);
   } /* endif */

   if (memcmp(aidname, PEN_STR, sizeof(PEN_STR) - 1) == 0)
   {
      strcpy(cicsname, DFHPEN_STR);
   } /* endif */

   if (memcmp(aidname, OPID_STR, sizeof(OPID_STR) - 1) == 0)
   {
      strcpy(cicsname, DFHOPID_STR);
   } /* endif */

   if (memcmp(aidname, TRIG_STR, sizeof(TRIG_STR) - 1) == 0)
   {
      strcpy(cicsname, DFHTRIG_STR);
   } /* endif */
}

/*
 * This routine inserts statements which simulate
 * a MAPFAIL condition.
 */

void insertHandleMapfail(char *ofsline, char *mapf)

{
   /* Write to the message file */
   writeErrorLine(outCounter, "HANDLE MAPFAIL statement inserted for EXEC CICS RECEIVE");

   /* insert an IF statement for the HANDLE AID */
   fprintf(newpgm,"%11sIF USRMAPLN IS EQUAL TO 0\n",ofsline);
   fprintf(newpgm,"%11s    GO TO %s\n",ofsline,mapf);
   fprintf(newpgm,"%11sEND-IF\n",ofsline);

   /* count output lines */
   outCounter += 3;
}

/*
 * This routine inserts statements which simulate
 * a handle aid.
 */

void insertHandleAid(char *ofsline, char *handleaids)

{
   char *aidname;
   char *aidval;
   char *endval;
   char varval[MAX_NAME_LEN];
   char cicsname[MAX_NAME_LEN];

   /* Write to the message file */
   writeErrorLine(outCounter, "HANDLE AID statements inserted for EXEC CICS RECEIVE");

   /* initialize pointer */
   aidname = handleaids;

   /* run through all the aid requests */
   while (1)
   {
      if (*aidname != '&')
      {
         break;
      } /* endif */

      /* point to the variable name */
      aidname++;

      /* initialize value pointer */
      aidval = aidname + 1;

      while ((*aidval != '='))
      {
         aidval++;
      } /* endwhile */

      /* point to the value */
      aidval++;

      /* find the next key name */
      endval = aidval;
      while ((*endval != '&') && (*endval != '\0'))
      {
         endval++;
      } /* endwhile */

      /* get the variable name for the GO TO */
      memset(varval, '\0', sizeof(varval));
      memcpy(varval, aidval, endval - aidval);

      /* build the first new line */
      cicsAIDname(aidname, cicsname);

      /* insert an IF statement for the HANDLE AID */
      fprintf(newpgm,"%11sIF USRAID IS EQUAL TO %s\n",ofsline,cicsname);
      fprintf(newpgm,"%11s    GO TO %s\n",ofsline,varval);
      fprintf(newpgm,"%11sEND-IF\n",ofsline);
      outCounter += 3;  /* count output lines */

      /* go on to the next entry */
      aidname = endval;
   } /* endwhile */
}

/*************************************************************/
/*                                                           */
/* This routine builds an in memory table with entries for   */
/* all SEND and RECEIVE commands found in the program.       */
/* The table is used to generate the appropriate entries     */
/* in the new 3270 front end program.                        */
/*                                                           */
/*************************************************************/

void BMSentry(BMSTABLE *BMStab,
			  char BMStype, 
			  const char *map, 
			  const char *mapset, 
			  char erasetype, 
			  char alarmtype,
			  char sendtype,
			  const char *frset,
			  const char *from,
			  const char *length,
			  const char *freekb)

{
	char mapname[MAX_LINE_LEN];		/* MAP parameter value    */
	char mapsetname[MAX_LINE_LEN];	/* MAPSET parameter value */
	char fromval[MAX_LINE_LEN];		/* FROM parameter value   */
	char lenval[MAX_LINE_LEN];		/* length parameter value */
	int i = 0;

	/* find the next unused entry */
	while ((BMStab[i].sendRecv != '\0') && (i <= MAX_BMS))
	{
		i++;
	}

	if (i <= MAX_BMS)
	{
		/* Fill in the command type, erase parameter and send parameter types */
		BMStab[i].sendRecv = BMStype;
		BMStab[i].eraseSW  = erasetype;
		BMStab[i].alarmSW  = alarmtype;
		BMStab[i].sendType = sendtype;

		/* Check for FREEKB parameter */
		if (freekb != NULL)
		{
			totalFreekb++;
			BMStab[i].freeKB    = 'F';
		}

		/* Check for FRSET parameter */
		if (frset != NULL)
		{
			totalFrset++;
			BMStab[i].frSet    = 'F';
		}

		/* Check if MAP parameter was specified */
		if (map != NULL)
		{
			/* Get the value of the MAP parameter */
			getValue(map, mapname);

			/* Copy it into the table entry */
			strncpy(BMStab[i].mapName, mapname, sizeof(BMStab[i].mapName));
		}

		if (length != NULL)
		{
			/* get value of LENGTH parameter  */
			getValue(length, lenval);

			/* Copy it into the table entry */
			strncpy(BMStab[i].dataLen, lenval, sizeof(BMStab[i].dataLen));
		} /* endif */

		if (mapset != NULL)
		{
			/* Get the value of the MAPSET parameter */
			getValue(mapset, mapsetname);

			/* Copy it into the table entry*/
			strncpy(BMStab[i].mapSet, mapsetname, sizeof(BMStab[i].mapSet));
		}

		if (from != NULL)
		{
			/* Get the value of the FROM/INTO parameter */
			getValue(from, fromval);

			/* Copy it into the table entry*/
			strncpy(BMStab[i].dataArea, fromval, sizeof(BMStab[i].dataArea));
		}
	}
}

/*
 * This routine handles an EXEC CICS RECEIVE command.
 */

void handleReceive(char *parmName,
                   char *ofsline,
                   char periodch,
                   char *handleaids,
                   char *mapf)

{
   char         insline[80];   /* line to be inserted          */

   /* build a new line to insert */
   sprintf(insline, "%11sIF %s IS NOT EQUAL TO USRMAPNAME",
           ofsline,parmName);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* build a new line to insert */
   sprintf(insline, "%11s    GO TO BADMAPNAME",ofsline);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* build a new line to insert */
   sprintf(insline, "%11sEND-IF%c",ofsline,periodch);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* increment the line counter for the output program */
   outCounter += 3;

   /* set indicator to add routine to end of program */
   badMap = 1;

   /* check if found HANDLE AID */
   if (strlen(handleaids) > 0)
   {
      insertHandleAid(ofsline, handleaids);
   } /* endif */

   /* check if found HANDLE CONDITION MAPFAIL */
   if (strlen(mapf) > 0)
   {
      insertHandleMapfail(ofsline, mapf);
   } /* endif */
}

/*
 * This routine handles an EXEC CICS SEND with the ERASE option
 */

void sendErase(char *ofsline, char erasetype, char periodch)

{
   char         insline[80];   /* line to be inserted          */

   /* build a new line to insert */
   sprintf(insline, "%11sMOVE '%c' TO ERASESW%c",
           ofsline,erasetype,periodch);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* increment the line counter for the output program */
   outCounter++;
}

/*
 * This routine handles an EXEC CICS SEND with the ERASE option
 */

void sendAlarm(char *ofsline, char alarmtype, char periodch)

{
   char         insline[80];   /* line to be inserted          */

   /* build a new line to insert */
   sprintf(insline, "%11sMOVE '%c' TO ALARMSW%c",
           ofsline,alarmtype,periodch);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* increment the line counter for the output program */
   outCounter++;
}

/*
 * This routine handles an EXEC CICS SEND MAP command.
 * It moves the map name to the USRMAPN field, and the
 * map set to the USRMAPS field.
 */

void handleSend(char *parm1, char *parm2, char *ofsline, char periodch)
{
	char insline[80];   /* line to be inserted          */

	/* build a new line to insert */
	sprintf(insline, "%11sMOVE %s TO USRMAPNAME%c",
			ofsline,parm1,periodch);

	/* write line to new program file */
	fprintf(newpgm,"%-80s\n",insline);

	/* increment the line counter for the output program */
	outCounter ++;

   if (strlen(parm2) > 0)
   {
		/* build a new line to insert */
		sprintf(insline, "%11sMOVE %s TO USRMAPSET%c",
				ofsline,parm2,periodch);

		/* write line to new program file */
		fprintf(newpgm,"%-80s\n",insline);

		/* increment the line counter for the output program */
		outCounter ++;
   }
}

/*
 * This routine handles an EXEC CICS SEND TEXT command.
 */

void sendText(char *p1,
              char *p2,
              char *p3,
              char *ofsline,
              char periodch)

{
   char         insline[80];   /* line to be inserted          */

   /* write out a line with a literal map name of SENDTEXT */
   handleSend("\'SENDTEXT\'", "", ofsline, periodch);

   /* build a new line to insert */
   sprintf(insline, "%11sMOVE %s TO TEXTAREA%c",
           ofsline,p1,periodch);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* build a new line to insert */
   sprintf(insline, "%11sMOVE %s TO USRMAPLN%c",
           ofsline,p2,periodch);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* increment the line counter for the output program */
   outCounter += 2;
}

/*
 * This routine handles an EXEC CICS SEND CONTROL command.
 */

void sendControl(char *ofsline, char periodch)

{
   char         insline[80];   /* line to be inserted          */

   /* build a new line to insert */
   sprintf(insline, "%11s%c", ofsline, periodch);

   /* write line to new program file */
   fprintf(newpgm,"%-80s\n",insline);

   /* increment the line counter for the output program */
   outCounter++;
}

 /*
  * Routine to handle EXEC CICS statements.
  * EXEC CICS statements can occupy one or more lines.
  * The routine scans for an END-EXEC, reading lines
  * until one is found, to a maximum of ten lines.
  * If 10 lines are read without finding an END-EXEC,
  * then an error message is generated and the lines
  * are written unchanged to the new file.
  */

int handleExec(char *firstLine,
               char *handleaids,
			   BMSTABLE *BMStab,
               char *mapf)

{
   int          linecount=0;
   int          i;
   int          saveCount;
   char         *origOfs;   /* original offset of command   */
   /* pointers to EXEC CICS commands and parameters */
   char         *receive;  /* pointer to RECEIVE command    */
   char         *send;     /* pointer to SEND command       */
   char         *converse; /* pointer to CONVERSE command   */
   char         *text;     /* pointer to TEXT parameter     */
   char         *control;  /* pointer to CONTROL parameter  */
   char         *accum;    /* pointer to ACCUM parameter    */
   char         *buffer;   /* pointer to BUFFER parameter   */
   char         *page;     /* pointer to PAGE parameter     */
   char         *convid;   /* pointer to CONVID parameter   */
   char         *session;  /* pointer to SESSION parameter  */
   char         *into;     /* pointer to INTO parameter     */
   char         *from;     /* pointer to FROM parameter     */
   char			*ctlchar;  /* pointer to CTLCHAR parameter  */
   char         *map;      /* pointer to MAP name parameter */
   char         *mapset;   /* pointer to MAPSET parameter   */
   char         *maponly;  /* pointer to MAPONLY parameter  */
   char			*dataonly; /* pointer to DATAONLY parameter */
   char         *erase;    /* pointer to ERASE parameter    */
   char         *eraseaup; /* pointer to ERASEAUP parameter */
   char			*freekb;   /* pointer to FREEKB parameter   */
   char			*frset;    /* pointer to FRSET parameter    */
   char			*alarm;    /* pointer to ALARM parameter    */
   char			*route;    /* pointer to ROUTE parameter    */
   char			*wait;     /* pointer to WAIT parameter     */
   char         *cursor;   /* pointer to CURSOR parameter   */
   char         *ret;      /* pointer to RETURN             */
   char         *address;  /* pointer to ADDRESS parameter  */
   char         *tctua;    /* pointer to TCTUA parameter    */
   char         *transid;  /* pointer to TRANSID parameter  */
   char         *commarea; /* pointer to COMMAREA parameter */
   char         *length;   /* pointer to LENGTH parameter   */
   char         *handle;   /* pointer to HANDLE parameter   */
   char         *aid;      /* pointer to AID parameter      */
   char         *mapfail;  /* pointer to MAPFAIL parameter  */
   char         *assign;   /* pointer to ASSIGN parameter   */
   char         *termcode; /* pointer to TERMCODE parameter */
   char         *link;     /* pointer to LINK               */
   char         *xctl;     /* pointer to XCTL               */
   char			*program;  /* pointer to PROGRAM parameter  */
   char         *endexec;  /* pointer to END-EXEC           */

   char         *lines[20];
   /* parameter values from EXEC CICS commands */
   char         p1[255];       /* parameter value              */
   char         p2[255];       /* parameter value              */
   char         p3[255];       /* parameter value              */
   char         *period;       /* pointer to period or NULL    */
   unsigned char periodch;     /* working variable - character */
   /* working storage */
   char         ch;            /* character variable */
   char			erasetype;	   /* Erase type character         */
   char			alarmtype;	   /* Alarm type character         */
   char			sendtype;      /* Type of BMS send             */
   char         command[2550]; /* consolidated command buffer  */
   char         insline[80];   /* line to be inserted          */
   char         ofsline[80];   /* line with origOFS blanks     */
   char			msgline[255];  /* message work area            */
   char         linktype[5];   /* LINK/XCTL work area          */
   char			temptext[255]; /* work area to build message   */

   /* Increase the EXEC CICS count */
   totalExecCICS++;

   saveCount = outCounter;  /* remember the first line number */
   lines[0] = firstLine;
   origOfs = findKeyword(firstLine, EXEC_STR);  /* get original offset */

   ofsline[0] = '\0';
   while ((strlen(ofsline) < (unsigned int) (origOfs - firstLine)) &&
          (strlen(ofsline) < 32))
   {
      strcat(ofsline, " ");
   }

   while (1)
   {
      endexec = findKeyword((char *)lines[linecount],ENDEXEC_STR);
      if (endexec != NULL)
      {
         break;
      } /* endif */

     /* go on to the next line */
     linecount++;

     if (linecount == 20)
     {
        /* More than 20 lines without END-EXEC */
        /* Write out the lines */
        for (i=0; i < linecount; i++)
        {
           fprintf(newpgm,"%s\n",(char *) lines[i]);
		   outCounter++;
        } /* endfor */

        return 50;
     } /* endif */

     lines[linecount] = malloc(255);
     if (fgets((char *)lines[linecount],255,cobolpgm) == NULL)
     {
        /* unexpected end of file */
        return 200;
     } /* endif */

      /* get rid of the newline character */
      i = strlen((char *)lines[linecount]) - 1;
          if (*(lines[linecount] + i) == '\n')
      {
         *(lines[linecount] + i) = '\0';
      }

    /* increment the line counter */
     inCounter++;
   } /* endwhile */

   /* we now have all of the lines for the EXEC CICS        */
   /* command - concatenate them into one contiguous string */
   command[0] = '\0';
   for (i=0; i <= linecount; i++)
   {
      /* bypass comment lines */
      if (*(lines[i] + 7) != '*')
      {
         strcat(command,(char *) (lines[i] + 7));
      } /* endif */
   } /* endfor */

   /* translate to upper case */
   strupr(command);

   /* check for EIB variables that must be changed */
   handleEIB(command);

   /* search for various keywords */
   receive  = findKeyword(command, RECEIVE_STR);
   send     = findKeyword(command, SEND_STR);
   converse = findKeyword(command, CONVERSE_STR);
   map      = findKeyword(command, MAP_STR);
   text     = findKeyword(command, TEXT_STR);
   control  = findKeyword(command, CONTROL_STR);
   accum    = findKeyword(command, ACCUM_STR);
   buffer   = findKeyword(command, BUFFER_STR);
   page     = findKeyword(command, PAGE_STR);
   convid	= findKeyword(command, CONVID_STR);
   session	= findKeyword(command, SESSION_STR);
   into     = findKeyword(command, INTO_STR);
   from     = findKeyword(command, FROM_STR);
   ctlchar  = findKeyword(command, CTLCHAR_STR);
   mapset   = findKeyword(command, MAPSET_STR);
   maponly  = findKeyword(command, MAPONLY_STR);
   dataonly = findKeyword(command, DATAONLY_STR);
   freekb   = findKeyword(command, FREEKB_STR);
   wait     = findKeyword(command, WAIT_STR);
   frset    = findKeyword(command, FRSET_STR);
   alarm    = findKeyword(command, ALARM_STR);
   route    = findKeyword(command, ROUTE_STR);
   erase    = findKeyword(command, ERASE_STR);
   eraseaup = findKeyword(command, ERASEAUP_STR);
   cursor   = findKeyword(command, CURSOR_STR);
   ret      = findKeyword(command, RETURN_STR);
   transid  = findKeyword(command, TRANSID_STR);
   commarea = findKeyword(command, COMMAREA_STR);
   length   = findKeyword(command, LENGTH_STR);
   address  = findKeyword(command, ADDRESS_STR);
   tctua    = findKeyword(command, TCTUA_STR);
   handle   = findKeyword(command, HANDLE_STR);
   aid      = findKeyword(command, AID_STR);
   mapfail  = findKeyword(command, MAPFAIL_STR);
   assign   = findKeyword(command, ASSIGN_STR);
   termcode = findKeyword(command, TERMCODE_STR);
   link     = findKeyword(command, LINK_STR);
   xctl     = findKeyword(command, XCTL_STR);
   program  = findKeyword(command, PROGRAM_STR);
   endexec  = findKeyword(command, ENDEXEC_STR);

   /* check if there is a period at the end of the statement */
   if (endexec != NULL)
   {
      period = strchr(endexec, '.');
   }
   else
   {
      period = strchr(command, '.');
   }

   /* check if we had a period after the END-EXEC */
   if (period == NULL)
   {
      periodch = ' ';
   }
   else
   {
      periodch = '.';
   }

   /* comment out the lines if necessary */
   if ((receive != NULL) || (send != NULL) ||
       (converse != NULL) || (ret != NULL) ||
       ((handle != NULL) && (aid != NULL)))
   {
	   /* Increase the total send/receive counter */
	   totSendRecv++;

      for (i=0; i <= linecount; i++)
      {
         (*(lines[i] + 6) = '*');
      } /* endfor */
   } /* endif */

   /* write the lines back out */
   for (i=0; i <= linecount; i++)
   {
      /* write line to new program file */
      fprintf(newpgm,"%s\n",(char *) lines[i]);

      /* increment the line counter for the output program */
      outCounter++;

      if (i > 0)
      {
         free(lines[i]);
      }
   } /* endfor */

   /* Set type fields to blank */
   erasetype = ' ';
   sendtype  = ' ';
   alarmtype = ' ';

   /* Check for various BMS command parameters */
   if (erase != NULL)
   {
	   totalErase++;
	   erasetype = 'E';
   } /* endif */
   
   /* Check for various BMS command parameters */
   if (alarm != NULL)
   {
	   totalAlarm++;
	   alarmtype = 'A';
   } /* endif */
   
   if (eraseaup != NULL)
   {
	   totalEraseAUP++;
	   erasetype = 'U';
   } /* endif */
   
   if (maponly != NULL)
   {
	   totalMaponly++;
	   sendtype = 'M';
   } /* endif */
   
   if (dataonly != NULL)
   {
	   totalDataonly++;
	   sendtype = 'D';
   } /* endif */

   /* Check for EXEC CICS RECEIVE command */
   if (receive != NULL)
   {
      if (map != NULL)
      {
		 /* Increment the counter */
		 totalRecvMap++;

         getValue(map, p1); /* get value of MAP parameter */
         handleReceive(p1, ofsline, periodch, handleaids, mapf);

	      /* handle EXEC CICS RECEIVE */
		 sprintf(msgline, "CICS RECEIVE MAP for map %s handled", p1);
		 writeErrorLine(saveCount, msgline);

         /* Build a BMS table entry to remember the BMS command */
	     BMSentry(BMStab, 'R', map, mapset, erasetype, alarmtype, sendtype, frset, into, length, freekb);
      }
      else
      {
         if (buffer != NULL)
		 {
			totalBuffer++;
            writeErrorLine(saveCount, "*** EXEC CICS RECEIVE BUFFER found");
		 }
		 else
		 {
            writeErrorLine(saveCount, "*** EXEC CICS RECEIVE with no MAP parameter");
		 }
      }
   }

   /* Check for EXEC CICS SEND command */
   if (send != NULL)
   {
	  if ((erase != NULL) || (eraseaup != NULL))
	  {
		sendErase(ofsline, erasetype, periodch);
	  }

	  if (alarm != NULL)
	  {
		sendAlarm(ofsline, alarmtype, periodch);
	  }

      if (page != NULL)
      {
		  totalPage++;
         writeErrorLine(saveCount, "*** SEND PAGE found");
      }

      if (map != NULL)
      {
         getValue(map, p1);    /* get value of MAP parameter     */
         getValue(mapset, p2); /* get value of MAPSET parameter  */
         handleSend(p1, p2, ofsline, periodch);

		 /* Increment the counter */
		 totalSendMap++;

         /* handle EXEC CICS SEND MAP */
         sprintf(msgline, "CICS SEND MAP handled for map %s mapset %s", p1, p2);
         writeErrorLine(saveCount, msgline);

         if (accum != NULL)
         {
			 totalAccum++;
            writeErrorLine(saveCount, "*** SEND ACCUM found");
         }

		 /* Build a BMS table entry to remember the BMS command */
		 BMSentry(BMStab, 'S', map, mapset, erasetype, alarmtype, sendtype, frset, from, length, freekb);
      }
      else
      {
         if (text != NULL)
         {
			 totalSendText++;
            /* handle EXEC CICS SEND TEXT */
            writeErrorLine(saveCount, "CICS SEND TEXT handled");

			/* get parameter values */
            getValue(from, p1);
            getValue(length, p2);
            getValue(cursor, p3);

            sendText(p1, p2, p3, ofsline, periodch);

		   /* Build a BMS table entry to remember the BMS command */
		   BMSentry(BMStab, 'T', "SENDTEXT", NULL, erasetype, alarmtype, sendtype, frset, from, length, freekb);
         }
         else
         {
            if (control != NULL)
            {
				totalControl++;
               /* handle EXEC CICS SEND CONTROL */
               writeErrorLine(saveCount, "*** CICS SEND CONTROL found");

               sendControl(ofsline, periodch);
            }
            else
            {
				if ((convid != NULL) || (session != NULL))
				{
					totalConvid++;
					writeErrorLine(saveCount, "*** APPC SEND found - ignored");
				}
				else
				{
					if ((ctlchar != NULL) || (from != NULL))
					{
						/* Treat a SEND FROM as a SEND TEXT FROM */
						if (from != NULL)
						{
							totalSendText++;
							/* handle EXEC CICS SEND TEXT */
							writeErrorLine(saveCount, "CICS SEND FROM handled as SEND TEXT");

							/* get parameter values */
							getValue(from, p1);
							getValue(length, p2);
							getValue(cursor, p3);

							sendText(p1, p2, p3, ofsline, periodch);

							/* Build a BMS table entry to remember the BMS command */
							BMSentry(BMStab, 'T', "SENDTEXT", NULL, erasetype, alarmtype, sendtype, frset, from, length, freekb);
						}
						else
						{
							totalSendTerm++;
							writeErrorLine(saveCount, "Send CTLCHAR found - removed");
						}
					}
					else
					{
						totalSendNoparm++;
						writeErrorLine(saveCount, "SEND found - no recognized parameters - removed");
					}
				}
            }
         }
      }
   }

   /* Check for BMS ROUTE command */
   if (route != NULL)
   {
	   totalRoute++;
      writeErrorLine(saveCount, "*** BMS ROUTE command found");
   }

   if (converse != NULL)
   {
	   totalConverse++;
      /* handle EXEC CICS CONVERSE */
      /* this requires program logic changes to allow the */
      /* program to operate in a call-return environment. */
      writeErrorLine(saveCount, "CONVERSE removed - program logic must be checked");

      /* handle as a send for now */
      if (map != NULL)
      {
         getValue(map, p1);    /* get value of MAP parameter    */
         getValue(mapset, p2); /* get value of MAPSET parameter */
		 /* Insert the statements to replace the EXEC CICS SEND MAP */
         handleSend(p1, p2, ofsline, periodch);
      }
      else
      {
         writeErrorLine(saveCount, "MAP parameter missing");
      }
   }

   /* handle EXEC CICS ADDRESS TCTUA or COMMAREA */
   if (address != NULL)
   {
	   totalAddress++;
      if (tctua != NULL)
      {
         writeErrorLine(saveCount, "CICS ADDRESS TCTUA handled");

         /* check for both commarea and tctua */
         if (commarea == NULL)
         {
            /* need to put period here */
            ch = periodch;
         }
         else
         {
            /* no need to put period here */
            ch = ' ';
         } /* endif */

         /* get value of TCTUA parameter */
         getValue(tctua, p1);
         sprintf(insline,"%11sSET %s TO ADDRESS OF USRTCTUA%c",
                 ofsline, p1, ch);
         fprintf(newpgm,"%-80s\n",insline);

         /* increment the line counter for the output program */
         outCounter++;
      }

      if (commarea != NULL)
      {
         writeErrorLine(saveCount, "CICS ADDRESS COMMAREA handled");

         /* get value of COMMAREA parameter */
         getValue(commarea, p1);
         sprintf(insline,"%11sSET %s TO ADDRESS OF %s%c",
                 ofsline, p1, DFHCOMM_NEW, periodch);
         fprintf(newpgm,"%-80s\n",insline);

         /* increment the line counter for the output program */
         outCounter++;
      }
   }

   if ((handle != NULL) && (mapfail != NULL))
   {
	   totalMapfail++;

      /* handle EXEC CICS HANDLE CONDITION MAPFAIL */
      writeErrorLine(saveCount, "*** CICS HANDLE CONDITION MAPFAIL found");

      /* remember that we found a HANDLE MAPFAIL */
      getValue(mapfail + 7, mapf);
   }

   if ((handle != NULL) && (aid != NULL))
   {
      /* handle EXEC CICS HANDLE AID */
      writeErrorLine(saveCount, "*** CICS HANDLE AID found");

      /* clear the table */
      *handleaids = '\0';

      /* remember that we found a HANDLE AID */
      processHandleAID(aid,ofsline,handleaids,periodch,saveCount);
   }

	/* Check for EXEC CICS LINK or XCTL */
	if ((link != NULL) || (xctl != NULL))
	{
		/* figure out what type of call this is */
		if (link != NULL)
		{
			strcpy(linktype, "LINK");
			totalLink++;
		}
		else
		{
			strcpy(linktype, "XCTL");
			totalXctl++;
		}

		/* get the name of the LINKED to PROGRAM into P1*/
		getValue(program, p1);

		/* build the message text */
		sprintf(msgline, "EXEC CICS %s to program %s found", linktype, p1);

		/* write a message to the log to document the program reference */
		writeErrorLine(saveCount, msgline);
	}

	/* Check for EXEC CICS ASSIGN TERMCODE */
	if ((assign != NULL) && (termcode != NULL))
	{
		/* Increment the statement counter */
		totalTermcode++;

		/* Issue an error message */
		writeErrorLine(saveCount, "ASSIGN TERMCODE found - program logic should be checked");
	}

	/* handle EXEC CICS RETURN */
	if (ret != NULL)
	{
		/* Increment return counter */
		totalReturn++;

		/* start to build message */
		strcpy(msgline, "EXEC CICS RETURN ");

		if (transid != NULL)
		{
			/* get value of TRANSID parameter */
			getValue(transid, p1);
			sprintf(insline,"%11sMOVE %s TO NXTRNID",ofsline,p1);
			fprintf(newpgm,"%-80s\n",insline);

			/* increment the line counter for the output program */
			outCounter++;

			/* Add transid to message */
			sprintf(temptext, "TRANSID(%s) ", p1);
			strcat(msgline, temptext);
		}

		if (commarea != NULL)
		{
			/* get value of COMMAREA parameter */
			getValue(commarea, p1);
			sprintf(insline,"%11sMOVE %s TO %s",ofsline,p1,DFHCOMM_NEW);
			fprintf(newpgm,"%-80s\n",insline);

			/* increment the line counter for the output program */
			outCounter++;

			/* Add transid to message */
			sprintf(temptext, "COMMAREA(%s) ", p1);
			strcat(msgline, temptext);
		}

		if (length != NULL)
		{
			/* get value of LENGTH parameter */
			getValue(length, p1);

			/* Add transid to message */
			sprintf(temptext, "LENGTH(%s) ", p1);
			strcat(msgline, temptext);
		}
		else
		{
			/* Zero out the user COMMAREA */
			strcpy(p1,"ZEROS");
		}

		sprintf(insline,"%11sMOVE %s TO USRCALEN",ofsline,p1);
		fprintf(newpgm,"%-80s\n",insline);

		/* increment the line counter for the output program */
		outCounter++;

		/* finish the message and write it out */
		strcat(msgline, "handled");

		if (length == NULL);
		{
			/* Note this fact on the message log */
			strcat(msgline, " User COMMAREA Length was set to zero");
		}

		/* Write the message to the message log */
		writeErrorLine(saveCount, msgline);

		/* write line to new program file */
		sprintf(insline,"%11sEXEC CICS RETURN END-EXEC%c",ofsline,periodch);
		fprintf(newpgm,"%-80s\n",insline);

		/* increment the line counter for the output program */
		outCounter++;
	}
}

void writeBadMapParagraph()
{
	fprintf(newpgm,"       BADMAPNAME.\n");
	fprintf(newpgm,"           MOVE \'MAP NAME DOES NOT MATCH\' TO TEXTAREA.\n");
	fprintf(newpgm,"           MOVE \'SENDTEXT\' TO USRMAPNAME.\n");
	fprintf(newpgm,"           MOVE ZEROS TO USRCALEN.\n");
	fprintf(newpgm,"           EXEC CICS RETURN END-EXEC.\n");

	/* increment the line counter for the output program */
	outCounter += 5;
}

void printHelp()

{
   printf("\n   CICS COBOL Application Conversion Aid V1.2\n");
   printf("\n   Command usage is:\n\n");
   printf("     CBLCVRT cobol-file-name <new-file-name> <new-front-end-name>\n\n");
   printf("The cobol file name is assumed to have a file ");
   printf("extension of CCP.\n");
   printf("If not new file name is given, the converted program wil have\n");
   printf("a file extension of NEW.\n");
   printf("If the new front end name parameter is specified, then a\n");
   printf("new skeleton 3270 front end program will be built.");
}

int main(int argc, char *argv[])

{
   int  foundProgid=1;      /* PROGRAM-ID paragraph found indicator */
   int  foundProcDiv=0;     /* Procedure Division found indicator   */
   int  foundLinkSect=0;    /* Linkage Section found indicator      */
   int  foundHandleAID=0;   /* Found a HANDLE AID statement         */
   int  dontWrite;
   int	fileloop;
   int	i;
   int	foundNewName=0;		/* Indicator if new name on command line */
   char *procedure;         /* Location of PROCEDURE              */
   char *division;          /* Location of DIVISION               */
   char *linkage;           /* Location of LINKAGE                */
   char *section;           /* Location of SECTION                */
   char *exec;              /* Location of EXEC                   */
   char *cics;              /* Location of CICS                   */
   char *level01;           /* Location of 01                     */
   char *dfhcomm;           /* Location of DFHCOMMAREA            */
   char *progid;            /* Location of PROGRAM-ID             */
   char *copy;				/* Location of COPY statement         */
   char *end;				/* Location of END                    */
   char *program;			/* Location of PROGRAM                */
   char *inc;				/* Location of -INC statement         */
   char *fileExt;           /* file extension                     */
   char *tempstr;
   char newProgName[MAX_PROG_NAME + 1]; /* Name of new program    */
   char frontProgName[MAX_PROG_NAME + 1]; /* Name of new 3270 front end */
   char cobolFileName[255]; /* File name of input program         */
   char newFileName[255];   /* File name of converted program     */
   char frontFileName[255]; /* File name of converted program     */
   char msgFileName[255];   /* File name of message file          */
   char copyName[255];      /* Name of copy book file             */
   char cobolLine[MAX_LINE_LEN]; /* Current COBOL Line            */
   char templine[MAX_LINE_LEN];  /* Work variable                 */
   char handleaids[1000];   /* save area for HANDLE AID           */
   char mapf[MAX_NAME_LEN]; /* save area for MAPFAIL              */
   BMSTABLE *BMStab;		/* Table of BMS commands found        */

/**************************************************************/
/*     Initialize variable areas                              */
/**************************************************************/

   memset(handleaids, '\0', sizeof(handleaids));
   memset(mapf, '\0', sizeof(mapf));
   memset(newFileName, '\0', sizeof(newFileName));
   memset(frontFileName, '\0', sizeof(frontFileName));

/**************************************************************/
/*     Allocate a table to remember BMS commands              */
/**************************************************************/

   BMStab = malloc((MAX_BMS + 1) * sizeof(BMSTABLE));
   memset(BMStab, '\0', (MAX_BMS + 1) * sizeof(BMSTABLE));

/**************************************************************/
/*     Initialize the file handle save stack                  */
/**************************************************************/

   for (i=0; i < MAX_COPY_DEPTH; i++)
   {
	   saveinput[i] = NULL;
	   saveoutput[i] = NULL;
   }

/**************************************************************/
/*     Parse the file name and generate new file name.        */
/**************************************************************/

   if (argc >= 2)        /* check for file name */
   {
      strcpy(cobolFileName,argv[1]);

      /* check for help request */
      if ((cobolFileName[0] == '?') || (cobolFileName[1] == '?'))
      {
         printHelp();
         exit(1);
      } /* endif */

	  /* check for optional new file name */
	  if (argc > 2)
	  {
		  /* Get the new file name */
		  strcpy(newFileName,argv[2]);

		  /* Make sure the file names are not the same */
		  if (strcmp(cobolFileName,newFileName) == 0)
		  {
			  /* clear the new file name */
			  newFileName[0] = '\0';

			  /* Print a message to let the user know */
			  printf("Old and new names are the same - ignoring new name");
		  }
		  else
		  {
			  processFilename(newFileName, newProgName);

			  /* If we have a new program name, make sure we search for the */
			  /*  PROGRAM-ID paragraph                                      */
			  if (strlen(newProgName) > 0)
			  {
				  /* Make sure we search for the PROGRAM-ID paragraph */
				  foundProgid = 0;
			  }
		}

	    if (argc > 3)
		{
		  /* Get the new 3270 front end file name */
		  strcpy(frontFileName,argv[3]);

		  /* Make sure the file names are not the same */
		  if ((strcmp(cobolFileName,frontFileName) == 0) ||
		      (strcmp(cobolFileName,frontFileName) == 0))
		  {
			  /* clear the new file name */
			  frontFileName[0] = '\0';

			  /* Print a message to let the user know */
			  printf("Old and new names are the same - ignoring new front end name");
		  }
		  else
		  {
			  /* Extract the program name from the file name */
			  processFilename(frontFileName, frontProgName);
		  }

		}
	  }
   }
   else
   {
      printHelp();
      exit(1);
   } /* endif */


/**************************************************************/
/* Generate the new file name                                 */
/**************************************************************/

   /* look for a period */
   fileExt = strchr(cobolFileName, '.');
   if (fileExt != NULL)
   {
      (*fileExt) = 0;
   } /* endif */

   /* Check if a new file name was specified on the command line */
   if (strlen(newFileName) == 0)
   {
	   /* remove the file extension if it exists */
	   strcpy(newFileName,cobolFileName);
	   strcat(newFileName,".new");
   }
   else
   {
	   /* Indicate specified on command line */
	   foundNewName = 1;
   }

   strcpy(msgFileName,cobolFileName);
   strcat(msgFileName,".msg");
   strcat(cobolFileName,".ccp");

/**************************************************************/
/* Open the cobol program, new file and message file.         */
/**************************************************************/

   if ((cobolpgm = fopen(cobolFileName, "r")) == NULL)
   {
      exit(100);
   }

   if ((newpgm = fopen(newFileName, "w")) == NULL)
   {
      fclose(cobolpgm);
      exit(100);
   }

   if ((msgfile = fopen(msgFileName, "w")) == NULL)
   {
      fclose(cobolpgm);
      fclose(newpgm);
      exit(100);
   }

/**************************************************************/
/*                                                            */
/* Routine to build the new COBOL program                     */
/*                                                            */
/*  This routine will read the COBOL program file, and        */
/*  search for 3270 terminal dependencies.  When found,       */
/*  the dependencies are removed or replaced.                 */
/*                                                            */
/**************************************************************/

   while (1)
   {
		/* read the next line in the source file */
		cobolLine[0] = '\0';
		fileloop = 1;
		while (fileloop == 1)
		{
			if (fgets(cobolLine, MAX_LINE_LEN, cobolpgm) == NULL)
			{
				if (saveinput[0] == NULL)
				{
					fileloop = 0;
				}
				else
				{
					/* close the current files */
					fclose(newpgm);
					fclose(cobolpgm);

					/* Write out the number of lines in the copy book */
					fprintf(msgfile, "  %d lines read in copy book\n", inCounter);
					fprintf(msgfile, "  %d lines written in copy book\n", outCounter);

					/* return to the previous files */
					restoreFiles();
				}
			}
			else
			{
				fileloop = 0;
			}
		}

		/* Did we read the next line? */
		if (cobolLine[0] == '\0')
		{
			break;
		}

      /* write the line to the new program */
      dontWrite = 0;

      /* increment the line counter */
	  inCounter++;

      /* get rid of the newline character */
      if (cobolLine[strlen(cobolLine) - 1] == '\n')
      {
         cobolLine[strlen(cobolLine) - 1] = '\0';
      }

      /* check for a comment */
      if ((strlen(cobolLine) > 6) && (cobolLine[6] != '*') && 
		  (cobolLine[6] != '/') && (cobolLine[6] != '-'))
      {
		  /* Check for a COPY book or -INC include statement */
          copy = findKeyword(cobolLine, COPY_STR);
		  inc = findKeyword(cobolLine, INC_STR);
		  if ((copy != NULL) || (inc != NULL))
		  {
			  if (copy != NULL)
			  {
				/* Increment the copy book count */
				totalCopy++;
			  }
			  else
			  {
				  /* Increment the include count */
				  totalInc++;

				  /* Set the copy pointer to the INC keyword */
				  /* from now on, use copy book processing */
				  copy = inc;
			  }

			  /* step past copy or -INC keyword */
			  copy += 5;

			  /* find the next non-blank character */
			  while (copy[0] == ' ')
			  {
				  copy++;
			  }

			  /* copy the rest of the line into the file name */
			  strcpy(copyName, copy);

			  /* find the end of the name */
			  tempstr = copyName;
			  while ((tempstr[0] != ' ') && (tempstr[0] != '\0') && (tempstr[0] != '.'))
			  {
				  tempstr++;
			  }

			  /* truncate the rest of the string after the name */
			  tempstr[0] = '\0';
			  openCopyBook(copyName, cobolLine, &dontWrite);
		  }

		 /* Search for the PROGRAM-ID paragraph */
		 if (foundProgid == 0)
		 {
			 progid = findKeyword(cobolLine, PROGRAMID_STR);
			 if (progid != NULL)
			 {
				 /* Remember we already found it */
				 foundProgid = 1;

				 /* Truncate the rest of the line */
				 progid[sizeof(PROGRAMID_STR) - 1] = '\0';

				 /* Finally, append the new program name */
				 strcat(progid, ". ");
				 strcat(progid, newProgName);
				 strcat(progid, ".");

				 /* Note the change in the message file */
		         writeErrorLine(outCounter, "PROGRAM-ID paragraph changed");
			 }
		 }

         if (foundProcDiv == 1)
         {
            exec = findKeyword(cobolLine, EXEC_STR);
			if (exec != NULL)
			{
				cics = findKeyword(exec + sizeof(EXEC_STR), CICS_STR);
			}

            if ((exec != NULL) && (cics != NULL))
            {
               handleExec(cobolLine, handleaids, BMStab, mapf);
               dontWrite = 1;
            }
            else
            {
               /* check for EIB variables that must be changed */
               handleEIB(cobolLine);
            } /* endif */

			/* Check for an END PROGRAM statement */
			end     = findKeyword(cobolLine, END_STR);
			program = findKeyword(cobolLine, PROGRAM_STR);
			if ((end != NULL) && (program != NULL) && (badMap == 1))
			{
				/* Write out the paragraph to handle map name mismatches */
				writeBadMapParagraph();

				/* Don't write it out again */
				badMap = 0;
			}
          } /* endif */

         /* Search for Linkage Section */
         linkage = findKeyword(cobolLine, LINKAGE_STR);
         section = findKeyword(cobolLine, SECTION_STR);
         if ((linkage != NULL) && (section != NULL))
         {
			 /* remember we found it */
            foundLinkSect = 1;

	        /* write LINKAGE SECTION line to new program file */
	        fprintf(newpgm,"%-80s\n",cobolLine);

			/* insert additional lines for DFHCOMMAREA */
			fprintf(newpgm,"       01  DFHCOMMAREA.\n");
			fprintf(newpgm,"           COPY STDHEAD.\n");
			fprintf(newpgm,"           02  USRCOMM.\n");
			fprintf(newpgm,"      * Insert User COMMAREA, TCTUA, BMS MAPS HERE\n");

			/* Increment the output line counter */
			outCounter += 5;

			/* Don't write the line again */
			dontWrite = 1;

  		    /* Note the change in the message file */
            writeErrorLine(outCounter - 2, "COPY for STDHEAD in DFHCOMMAREA inserted");
         }

		 /* Search for 01 DFHCOMMAREA */
		 level01 = strstr(cobolLine, LEVEL01_STR);
		 dfhcomm = findKeyword(cobolLine, DFHCOMM_STR);
		 if ((level01 != NULL) && (dfhcomm != NULL) && (foundProcDiv == 0))
		 {
			 /* Remove the 01 DFHCOMMAREA statement */
			 level01[0] = '\0';

			 /* find the period after DFHCOMMAREA */
			 dfhcomm += strlen(DFHCOMM_STR);
			 while ((dfhcomm[0] != '\0') && (dfhcomm[0] != '.'))
			 {
				 dfhcomm++;
			 }

			 /* remove the period */
			 if (dfhcomm[0] = '.')
			 {
				 /* Skip the period */
				 dfhcomm++;

				 /* Make a copy of the rest of the line */
				 strcpy(templine, dfhcomm);

				 /* Concatenate it to the end of the current line */
				 strcat(cobolLine, "   ");
				 strcat(cobolLine, templine);
			 }
		 }

         /* Search for PROCEDURE DIVISION */
         procedure = findKeyword(cobolLine,PROCEDURE_STR);
         division = findKeyword(cobolLine, DIVISION_STR);

         if ((procedure != NULL) && (division != NULL))
         {
            foundProcDiv = 1;
            if (foundLinkSect == 0)
            {
               /* Insert LINKAGE SECTION and      */
               /* DFHCOMMAREA statements.         */
               fprintf(newpgm,"       LINKAGE SECTION.\n");
               fprintf(newpgm,"       01 DFHCOMMAREA.\n");
               fprintf(newpgm,"           COPY STDHEAD.\n");

               writeErrorLine(outCounter,"LINKAGE SECTION INSERTED");

               outCounter += 3;
            } /* endif */
         }

      } /* endif procedure division */

      /* write out the line */
      if (dontWrite == 0)
      {
         /* write line to new program file */
         fprintf(newpgm,"%-80s\n",cobolLine);

         /* increment the line counter for the output program */
         outCounter++;
      }
   } /* endwhile */

   if (badMap == 1)
   {
		/* Write out the paragraph to handle map name mismatches */
		writeBadMapParagraph();
   } /* endif */

   fprintf(msgfile,"\n%5d Input lines read\n",inCounter);
   fprintf(msgfile,"%5d Output lines written\n",outCounter);
   fprintf(msgfile,"%5d Total copy statements found\n",totalCopy);
   fprintf(msgfile,"%5d Total -INC statements found\n",totalInc);
   /* print the other totals */
   fprintf(msgfile,"\n%5d EXEC CICS Statements\n",totalExecCICS);
   fprintf(msgfile,"%5d Total EXEC CICS LINK statements found\n",totalLink);
   fprintf(msgfile,"%5d Total EXEC CICS XCTL statements found\n",totalXctl);
   fprintf(msgfile,"\n%5d SEND MAP statements\n",totalSendMap);
   fprintf(msgfile,"%5d SEND TEXT statements\n",totalSendText);
   fprintf(msgfile,"%5d SEND CONTROL statements\n",totalControl);
   fprintf(msgfile,"%5d SEND CONVID or SESSION statements\n",totalConvid);
   fprintf(msgfile,"%5d SEND with FROM or CTLCHAR only\n",totalSendTerm);
   fprintf(msgfile,"%5d SEND with no other parameters\n",totalSendNoparm);
   fprintf(msgfile,"%5d RECEIVE MAP statements\n",totalRecvMap);
   fprintf(msgfile,"%5d RECEIVE BUFFER statementss\n",totalBuffer);
   fprintf(msgfile,"%5d CONVERSE statements\n",totalConverse);
   fprintf(msgfile,"%5d RETURN statements\n",totalReturn);
   fprintf(msgfile,"%5d ADDRESS statements\n",totalAddress);
   fprintf(msgfile,"%5d HANDLE MAPFAIL statements\n",totalMapfail);
   fprintf(msgfile,"%5d HANDLE AID statements\n",totalHandleAID);
   fprintf(msgfile,"%5d ROUTE statements\n",totalRoute);
   fprintf(msgfile,"\n%5d MAPONLY parameters\n",totalMaponly);
   fprintf(msgfile,"%5d DATAONLY parameters\n",totalDataonly);
   fprintf(msgfile,"%5d ERASE parameters\n",totalErase);
   fprintf(msgfile,"%5d ERASEAUP parameters\n",totalEraseAUP);
   fprintf(msgfile,"%5d ALARM parameters\n",totalAlarm);
   fprintf(msgfile,"%5d ACCUM parameters\n",totalAccum);
   fprintf(msgfile,"%5d PAGE parameters\n",totalPage);
   fprintf(msgfile,"%5d FRSET statements\n",totalFrset);
   fprintf(msgfile,"%5d FREEKB statements\n",totalFreekb);
   fprintf(msgfile,"\n%5d EIBAID variables\n",totEibaid);
   fprintf(msgfile,"%5d EIBTRMID variables\n",totEibtrmid);
   fprintf(msgfile,"%5d EIBTRNID variables\n",totEibtrnid);
   fprintf(msgfile,"%5d EIBCALEN variables\n",totEibcalen);
   fprintf(msgfile,"%5d EIBCPOSN variables\n",totEibcposn);
   fprintf(msgfile,"\n%5d DFHCOMMAREA references\n",totCommarea);
   fprintf(msgfile,"%5d ASSIGN TERMCODE references\n",totalTermcode);

   fclose(cobolpgm);
   fclose(newpgm);

/**************************************************************/
/* Build the optional new 3270 front end program              */
/**************************************************************/

   if ((strlen(frontFileName) > 0) && (totSendRecv > 0))
   {
	   if ((frontpgm = fopen(frontFileName, "w")) == NULL)
	   {
		   /* Can't open the file, log the error */
		   fprintf(msgfile,BAD_FRONTEND_FILE);
	   }
	   else
	   {
		   if ((prototyp = fopen(PROTOTYP_FILE, "r")) == NULL)
		   {
			   /* We tried the current directory, next try the */
			   /* program directory                            */
			   strcpy(newProgName, argv[1]);
			   while ((strlen(newProgName) > 0) && 
				      (newProgName[strlen(newProgName)] != '\\'))
			   {
				   newProgName[strlen(newProgName)] = '\0';
			   }

			   /* append the file name */
			   strcat(newProgName, PROTOTYP_FILE);
				
			   /* try the open again */
			   prototyp = fopen(PROTOTYP_FILE, "r");
		   }

		   if (prototyp == NULL)
		   {
			   /* Can't open the template file, log the error */
			   fprintf(msgfile,BAD_FRONTEND_FILE);
		   }
		   else
		   {
			   /* tell the user what we are doing */
			   fprintf(msgfile,CREATE_NEW_FRONT_END, 
					   frontProgName);

			   /* build a skeleton 3270 front end program */
			   build3270pgm(BMStab,frontProgName, newProgName);

			   /* close the prototype file */
			   fclose(prototyp);
		   }

		   /* close the new 3270 front end program file */
		   fclose(frontpgm);
	   }
   }

   /* Close the message file */
   fclose(msgfile);

   /* Be a good citizen and free the acquired storage */
   free(BMStab);

/**************************************************************/
/* Finished - return normally                                 */
/**************************************************************/

   return 0;
}

/*-------------*/
/* End Of File */
/*-------------*/
